#ifndef _GenericList_h
#define _GenericList_h

#include "global/global.h"
#include <dir.h>
#include <fstream.h>
#include <except.h>
#include <cstring.h>

#define GL_MINSIZE 		8
#define GL_RESIZEFACTOR 1.2

template<class T> class GenericList
{
   private:
		fstream*		 _p_stream;
      char			 _filename[MAXDIR];
      unsigned int _next_element, _size, _min_size;
      float			 _resize_factor;

	protected:
		bool			 _syncronized;      // sync. file and internal data
   	T* 			 _pArray;

   public:
   	GenericList(const char* pFilename = NULL, const unsigned int = GL_MINSIZE, const float = GL_RESIZEFACTOR) throw(xmsg);
   	~GenericList();

      unsigned int    Add(const T&);
      void			    Remove(const unsigned int);
      T&   			    operator[] (const unsigned int) throw(xmsg);
      unsigned int    GetNoElements() const {return _next_element;};
      void			    Empty() throw(xmsg);
		void			    Flush() throw(xmsg);
      void			    SetSyncronized(bool syncronized) {_syncronized = syncronized;};

   private:
		void CheckArraySize(const unsigned int);
}; // class GenericList

template<class T> class GenericSortableList : public GenericList<T>
{
   public:
   	enum Direction {SORT_ASCENDING, SORT_DESCENDING};

	protected:
		Direction _direction;

	public:
		GenericSortableList(const char* pFilename = NULL, const unsigned int = GL_MINSIZE, const float = GL_RESIZEFACTOR);

		void SetDirection(const Direction);
		void Sort();
}; // class GenericSortableList

template<class T> GenericList<T>::GenericList(const char* pFilename, const unsigned int MinSize = GL_MINSIZE,
															 const float ResizeFactor = GL_RESIZEFACTOR) throw(xmsg)
                                              : _min_size(MinSize), _resize_factor(ResizeFactor)
{
	streampos total_file_size;

   _pArray = 0;
	_size = _next_element = 0;
	_p_stream = NULL;

   if(pFilename)
   {
   	strcpy(_filename, pFilename);

		// read data from file if there is any

		fstream stream(pFilename, ios::in|ios::binary);

		if(!stream.bad())
      {
			stream.seekp(0, ios::end);
         total_file_size = stream.tellg();
			stream.seekp(0, ios::beg);

			T tmpT;

         SetSyncronized(false);

         while(stream.tellg() < total_file_size)
         {
            tmpT.Read(stream);
            Add(tmpT);
         } // while

         SetSyncronized(true);
         stream.close();
      } // if

		_p_stream = new fstream(pFilename, ios::in|ios::out|ios::binary|ios::app);
      if(_p_stream->bad()) throw(xmsg(string("File could not be opened in GenericList::GenericList()")));

	   SetSyncronized(true);
      return;
   } // if

   SetSyncronized(false);
} // GenericList::GenericList

template<class T> GenericList<T>::~GenericList()
{
   Flush();

   if(_pArray)
      delete[] _pArray;

   if(_p_stream) delete _p_stream; _p_stream = NULL;

   _pArray = 0; _next_element = _size = 0;
} // GenericList::~GenericList

template<class T> unsigned int GenericList<T>::Add(const T& Element)
{
   CheckArraySize(_next_element);

	_pArray[_next_element++] = Element;

   // add element to file

	if((_p_stream) && (_syncronized))
      Element.Write(*_p_stream);

   return _next_element-1;
} // GenericList::Add

template<class T> void GenericList<T>::Remove(const unsigned int position)
{
	T tmpT = (*this)[GetNoElements()-1];

	(*this)[position] = tmpT;
   _next_element--;

   // remove elements from file

	if((_p_stream) && (_syncronized))
   {
      _p_stream->seekp(T::SizeOf()*position, ios::beg);
      tmpT.Write(*_p_stream);
      _p_stream->seekp(T::SizeOf()*_next_element, ios::beg);
   } // if
} // GenericList::Remove

template<class T> T& GenericList<T>::operator[] (const unsigned int pos) throw(xmsg)
{
	if(pos >= _next_element) throw(xmsg(string("attempt to access an index out of range in GenericList<T>::operator[]")));


	return _pArray[pos];
} // GenericList::operator[]

template<class T> void GenericList<T>::CheckArraySize(const unsigned int size)
{
	T* pTmpList;

   if(_size < (size+1))
   {
      if(_size == 0) _size = _min_size;
      else _size *= _resize_factor;

      pTmpList = new T [_size];

      for(unsigned int i = 0; i < _next_element; i++)
         pTmpList[i] = _pArray[i];

		if(_pArray) delete[] _pArray;

		_pArray = pTmpList;
   } // if
} // GenericList::CheckArraySize

template<class T> void GenericList<T>::Empty() throw(xmsg)
{
	if(_pArray)
   	delete[] _pArray;

   _pArray = 0; _next_element = _size = 0;

   // delete data in file

   if(_p_stream)
   {
   	delete _p_stream;
      _p_stream = new fstream(_filename, ios::out|ios::binary);
      if(_p_stream->bad()) throw(xmsg(string("File could not be cleared in GenericList<T>::Empty()")));
   } // if
} // GenericList<T>::Empty

template<class T> void GenericList<T>::Flush() throw(xmsg)
{
	if(_p_stream)
   {
		delete _p_stream;

		_p_stream = new fstream(_filename, ios::out|ios::binary);
      if(_p_stream->bad()) throw(xmsg(string("couldn't open file in GenericList<T>::Flush() for output")));

		for(int i = 0; i < GetNoElements(); i++)
      	(*this)[i].Write(*_p_stream);

      delete _p_stream;
		_p_stream = new fstream(_filename, ios::in|ios::out|ios::binary|ios::app);
   } // if
} // GenericList<T>::Flush

template<class T> GenericSortableList<T>::GenericSortableList(const char* pFilename = NULL, const unsigned int min_size = GL_MINSIZE, const float resize_factor = GL_RESIZEFACTOR)
                   														  : GenericList<T>(pFilename, min_size, resize_factor)
{
	_direction = SORT_ASCENDING;
} // GenericSortableList<T>::GenericSortableList

template<class T> void GenericSortableList<T>::SetDirection(const Direction direction)
{
	_direction = direction;
} // GenericSortableList<T>::SetDirection

template<class T> void GenericSortableList<T>::Sort()
{
	bool changed;

   do
   {
      changed = false;

		for(int i = GetNoElements() - 1; i > 0; i--)
	   {
         if(((_direction == SORT_ASCENDING) && ((*this)[i] < (*this)[i-1])) || ((_direction == SORT_DESCENDING) && ((*this)[i] > (*this)[i-1])))
         {
            swap((*this)[i], (*this)[i-1]);
            changed = true;
         } // if
		} // for
	} while(changed == true);

   if(_syncronized) Flush();
} // GenericSortableList::Sort
#endif